`define OP  2:0 
`define	PC 11:4
`define	RD 14:12
`define	RA 18:16
`define	RB 22:20
`define	IM 35:24
`define OPHALT 3
`define OPLW 41
`define OPSW 42

module jsim3 
#(parameter DATA_WIDTH=16, parameter ADDR_WIDTH=5)
(
	input clk, reset_n,
	output [(ADDR_WIDTH-1):0] addr,
	output [2:0] ra, rb, rd,
	output  [(DATA_WIDTH-1):0] q, qa, qb, qd,
	output reg [(DATA_WIDTH-1):0] inst,
	//output reg [2:0] state, 
	output reg [47:0] IF,ID,EX,
	output reg [(DATA_WIDTH-1):0] ncycles
);


single_port_ram dev1(
	.data(EX_qb),
	.addr(addr),
	.we(we),
	.clk(clk),
	.q(q)
);

reg [(DATA_WIDTH-1):0] EX_qa, EX_qb;

regfile dev2(
	.clk(clk),
	.we(1'b1),
	.ra(ra),
	.qa(qa),
	.rb(rb),
	.qb(qb),
	.rd(rd),
	.qd(qd)
);

reg [4:0] pc;
//wire [(DATA_WIDTH-1):0] q;
wire [2:0] opcode;

// Instruction Fetch (IF)

wire [15:0] qi = (state==IFE? q: inst);
assign opcode = qi[15:13];

// opcodes	
`include "opcodes.txt"

wire op_add = opcode==ADD;
wire op_nand = opcode==NAND;
wire op_sw = opcode==SW;
wire op_lw = opcode==LW;
wire op_bne = opcode==BNE;
wire op_jalr = opcode==JALR;

assign flag1 = op_sw||op_bne;   // no write-back
assign flag2 = op_add||op_nand; // RRR format
wire [2:0] rd1 = flag1? 3'h0: qi[12:10];
assign ra = qi[9:7];
assign rb = flag2? qi[2:0]: (flag1? qi[12:10]: 3'h0);


reg [(ADDR_WIDTH-1):0] jump;

// EX Stage

wire we = (state==MEM) && EX[`OPSW];
wire memory_access = (EX[`OPSW] || EX[`OPLW]);

wire do_branch = op_bne && (qa!=qb);


wire [4:0] nextpc = pc + 1'b1;
wire [4:0] newpc = nextpc;
//wire [4:0] newpc = (op_jalr||do_branch? jump: nextpc);

assign addr = (state==MEM? d1: newpc );


wire [15:0] d1;

wire [4:0] ex_pc = EX[`PC] + 1'b1;

alu dev3(
  .a(EX_qa),
  .b(EX_qb),
  .pc({11'h000,ex_pc}),
  .im10(EX[`IM]),
  .d(d1),
  .opcode(EX[`OP])
);

assign rd = EX[`RD];
// writeback register multiplexer
assign qd = (EX[`OPLW]? q: d1);


reg [1:0] state;
localparam 
	IFE = 2'h0,
	MEM = 2'h1,
	HALT = 2'h2;

// IF[3] <==> halt
// IF[41] <==> op_lw;
// IF{42] <==> op_sw;

always @(posedge clk, negedge reset_n)
begin
  if (!reset_n)
  begin
   pc <= 5'h1F;
   IF <= 48'h0000_0000_0000;
   state <= IFE;
   end
  else 
	case (state)
	IFE:
	begin
		// IF stage
		IF[`OP] <= opcode;
		IF[`OPHALT] <= (qi==16'he071); // is halt
		IF[`PC] <= addr;
		IF[`RD] <= rd1;
		IF[`RA] <= ra;
		IF[`RB] <= rb;
		IF[`IM] <= qi[9:0];
		IF[`OPLW] <= op_lw;
		IF[`OPSW] <= op_sw;
		pc <= addr;
		// ID stage
		ID <= IF;
		// EX stage
		EX <= ID;
		EX_qa <= qa;
		EX_qb <= qb;
		if (memory_access) state <= MEM;
		//jump <= op_jalr? qa[(ADDR_WIDTH-1):0]: d1[(ADDR_WIDTH-1):0];
		if (EX[`OPHALT]) state <= HALT;
    end
    MEM:
    begin
		inst <= q;
		state <= IFE;
	end
	default:
		state <= HALT;
  endcase
end

always @(posedge clk or negedge reset_n)
begin
	if (!reset_n) ncycles <= 16'h0000;
	else if (state != HALT) ncycles <= ncycles + 1'b1;
end

endmodule
